// Copyright 2021 The Sigstore Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package cryptoutils

import (
	"testing"

	"github.com/google/go-cmp/cmp"
)

func TestPEMEncode(t *testing.T) {
	const testPemType PEMType = "ARBITRARY PEM TYPE"
	expected := []byte(`-----BEGIN ARBITRARY PEM TYPE-----
AAECAwQ=
-----END ARBITRARY PEM TYPE-----
`)
	b := []byte{0x00, 0x01, 0x02, 0x03, 0x04}
	pemBytes := PEMEncode(testPemType, b)
	if d := cmp.Diff(expected, pemBytes); d != "" {
		t.Errorf("MarshalCertificateToPEM() returned unexpected PEM (-want +got): %s", d)
	}
}
